/*
The MIT License (MIT)

Copyright (c) 2020 Charles Gunyon

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

void test_msgpack(void **state);
void test_fixedint(void **state);
void test_numbers(void **state);
void test_conversions(void **state);
void test_nil(void **state);
void test_boolean(void **state);
void test_bin(void **state);
void test_string(void **state);
void test_array(void **state);
void test_map(void **state);
void test_ext(void **state);
void test_obj(void **state);
#ifndef CMP_NO_FLOAT
void test_float_flip(void **state);
#endif
void test_skipping(void **state);
void test_deprecated_limited_skipping(void **state);
void test_errors(void **state);
void test_version(void **state);

/* vi: set et ts=2 sw=2: */
